#!/usr/bin/env python
"""
Tested with Python 2.4 to 3.6

Ancient Python compatibility is necessary as some Centreon running under
Centos may not have recent Python interpreters.
We also do not require any libraries, only some Net-SNMP commands which
should be easy to get on any monitoring system.
"""

from __future__ import division

import sys
import optparse
import subprocess
import re
import errno


class NagiosExitCode:
    """
    Nagios expects some defined exit codes from plugins
    """
    ok = 0
    warning = 1
    critical = 2
    unknown = 3


def die(message):
    sys.stderr.write(message + '\n')
    sys.stderr.write('You may use -h/--help to list the allowed options.\n')
    sys.exit(NagiosExitCode.unknown)


if sys.version_info < (2, 4):
    die("Expected at least Python 2.4")


parser = optparse.OptionParser()
parser.add_option('-H', '--hostname', dest='hostname',
                  metavar='<host>', help='FQDN or IP addres of the host to query')
parser.add_option('-p', '--port', dest='port', default=161,
                  metavar='<port>', help='SNMP port to use (default: %default)')
parser.add_option('-C', '--comunity', dest='community', default='public',
                  metavar='<community>', help='SNMPv2 community to use (default: %default)')
parser.add_option('-a', '--bca', dest='bca',
                  metavar='<BCA>', help='name of the BCA')
parser.add_option('-n', '--bcn', dest='bcn',
                  metavar='<BCN>', help="name of the BCN (example: 'Zone A - Zone B (capture)')")
parser.add_option('--bcnatob', action='store_true', dest='bcnatob',
                  help='only use the status of the zone A')
parser.add_option('--bcnbtoa', action='store_true', dest='bcnbtoa',
                  help='only use the status of the zone B')
parser.add_option('-m', '--mib', dest='mib', default='PERFORMANCEVISION-MIB',
                  metavar='<mib>', help='name of the PerformanceVision MIB (default: %default)')
parser.add_option('-M', '--mibsdir', dest='mib_directories',
                  default='+/usr/share/snmp/mibs/ietf:/usr/share/mibs/ietf:/usr/share/snmp/mibs',
                  metavar='<directories>', help=(
                      "colon separated list of diretories in which MIBs can be found "
                      "(use 'net-snmp-config --default-mibdirs' to list them, default: %default) "
                      "- on Debian systems, /usr/share/.../ietf provides SNMPv2 MIBs and can be "
                      "fetched via the download-mibs command from snmp-mibs-downloader package "
                      "(as libsnmp-base only provides a couple of MIBs in /usr/share/snmp/mibs) "
                      "- on CentOS systems, /usr/share/snmp/mibs provides SNMPv2 MIBs from "
                      "net-snmp-libs (if more MIBs were required, libsmi package provides "
                      "/usr/share/mibs)"))
options, args = parser.parse_args()


if options.hostname is None:
    die('Specify the remote host via -H/--hostname.')
if ((options.bcn is None and options.bca is None)
    or (options.bcn is not None and options.bca is not None)):
    die('Specify mode via either -a/--bca <BCA> or -n/--bcn <BCN>.')
if options.bcn is not None and options.bcnatob is not None and options.bcnbtoa is not None:
    die('Asked to only retrieve both BCN A to B (--bcnatob) and BCN B to A (--bcnbtoa) in BCN mode.')
if options.bca is not None and (options.bcnatob is not None or options.bcnbtoa is not None):
    die('Asked to only retrieve a BCN side (--bcnatob or --bcnbtoa) in BCA mode.')


def snmp_cmd(args):
    """
    SNMP commands wrapper, force some MIB-related options

    Returns a tuple (command, exit code, stdout)
    """
    args = list(args)
    args += ['-m', options.mib] # force our MIB
    if options.mib_directories is not None:
        args += ['-M', options.mib_directories] # force MIBs directories if specified

    try:
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        e = sys.exc_info()[1]
        if e.errno == errno.ENOENT:
            die('%r not found but is required.' % args[0])
    stdout, stderr = p.communicate()
    stdout, stderr = stdout.decode('utf-8'), stderr.decode('utf-8')
    sys.stderr.write(stderr) # always output stderr for debugging purposes (if any)
    return ' '.join(args), p.returncode, stdout, stderr


def snmp_translate(args):
    """
    Wrapper for snmptranslate
    """
    return snmp_cmd(['snmptranslate'] + args)


def snmp_walk(args):
    """
    Wrapper for snmpwalk
    """
    return snmp_cmd(['snmpwalk',
                     '-OsqU', # necessary formatting for easy parsing
                     '-v', '2c', # v2
                     '-c', options.community, # snmp scommunity
                     '%s:%s' % (options.hostname, options.port), # host
                    ] + args)


# try to find our OID so we are sure the MIB is installed
command, exit_code, stdout, stderr = snmp_translate(['-IR', 'performanceVision'])
if (exit_code != 0
    or 'Did not find' in stderr): # snmptranslate don't exit != 0 when it can't import stuff
    die("""PERFORMANCEVISION-MIB or its dependencies seem to be missing on your system.
To diagnose the problem, you may want to run %r yourself.""" % command)


class Mode:
    # map mib value to human readable form and exit code
    status = {
        'ok': ('is OK', NagiosExitCode.ok),
        'warning': ('is WARN', NagiosExitCode.warning),
        'alert': ('is CRIT', NagiosExitCode.critical),
        'na': ("isn't evaluated", NagiosExitCode.ok),
        'nothing': ("hasn't any data", NagiosExitCode.ok),
        'notEnough': ("hasn't enough data", NagiosExitCode.ok),
    }
    table_oid = None # the table's OID
    name_oid = None # the name to match
    status_oid = None # the status' OID
    # performance data, only elements present here will be shown
    perfdata_aliases = {}
    # map to nagios performance data thresholds
    perfdata_thresholds = {}
    perfdata_thresholds_percentages = []

    @classmethod
    def get_status_tuple(cls, results):
        if cls.status_oid not in results:
            die("Invalid SNMP restults")
        else:
            return cls.status[results[cls.status_oid]]

    @classmethod
    def get_status(cls, results):
        return cls.get_status_tuple(results)[0]

    @classmethod
    def get_exit_code(cls, results):
        return cls.get_status_tuple(results)[1]

    @classmethod
    def get_label(cls, results):
        """
        Returns the label to display
        """
        raise NotImplementedError()

    @classmethod
    def get_thresholds(cls, key, results):
        """
        Returns the min, max, warn, crit thresholds.
        Take care of converting to percentage if necessary.
        """
        def get_threshold(threshold, key, results):
            if key not in cls.perfdata_thresholds:
                return None

            key = cls.perfdata_thresholds[key].get(threshold)
            value = results.get(key)
            if value:
                try:
                    value = int(value)
                except ValueError:
                    die("Could not convert %r to integer for %r" % (value, key))
            return value

        thresholds = {}
        for threshold in ['min', 'max', 'warn', 'crit']: # order is important
            threshold_value = get_threshold(threshold, key, results)
            if threshold_value is None:
                continue
            # max has to be fetched before calculating the value from percentage for warn/crit
            if (key in mode.perfdata_thresholds_percentages
                and threshold in ('warn', 'crit')):
                assert 'max' in thresholds
                threshold_value = thresholds['max'] * threshold_value / 100
            thresholds[threshold] = threshold_value
        return thresholds


class BCA(Mode):
    query = options.bca
    table_oid = 'bcaStateTable'
    name_oid = 'bcaName'
    status_oid = 'bcaStatus'
    perfdata_aliases = {
        'bcaBandwidthClient': 'bcaBwClt',
        'bcaBandwidthServer': 'bcaBwSrv',
        'bcaDTTClient': 'bcaDTTClt',
        'bcaDTTServer': 'bcaDTTSrv',
        'bcaEURT': 'bcaEURT',
        'bcaRTTClient': 'bcaRTTClt',
        'bcaRTTServer': 'bcaRTTSrv',
        'bcaSRT': 'bcaSRT',
        'bcaThresholdAlert': 'bcaThrAlert',
        'bcaThresholdWarning': 'bcaThrWarn',
        'bcaTrafficClient': 'bcaTrafficClt',
        'bcaTrafficServer': 'bcaTrafficSrv',
    }
    perfdata_thresholds = {
        'bcaEURT': {
            'warn': 'bcaThresholdWarning',
            'crit': 'bcaThresholdAlert',
        },
        'bcaSRTCount': {
            'min': 'bcaThresholdMinSRTcount',
        }
    }

    @classmethod
    def get_label(cls, results):
        return '%s BCA %s' % (results['bcaName'], cls.get_status(results))


class BCN(Mode):
    query = options.bcn
    table_oid = 'bcnStateTable'
    name_oid = 'bcnName'
    if options.bcnatob:
        status_oid = 'bcnStatusAtoB'
    elif options.bcnbtoa:
        status_oid = 'bcnStatusBtoA'
    else:
        status_oid = 'bcnGlobalStatus'
    perfdata_aliases = {
        'bcnBandwidthAtoB': 'bcnBwAtoB',
        'bcnBandwidthBtoA': 'bcnBwBtoA',
        'bcnPacketsAtoB': 'bcnPktsAtoB',
        'bcnPacketsBtoA': 'bcnPktsBtoA',
        'bcnRRAtoB': 'bcnRRAtoB',
        'bcnRRBtoA': 'bcnRRBtoA',
        'bcnRTTAtoB': 'bcnRTTAtoB',
        'bcnRTTBtoA': 'bcnRTTBtoA',
        'bcnThresholdBandwAvailableAtoB': 'bcnThrBwAvlAtoB',
        'bcnThresholdBandwAvailableBtoA': 'bcnThrBwAvlBtoA',
        'bcnThresholdBandwMinAtoB': 'bcnThrBwMinAtoB',
        'bcnThresholdBandwMinBtoA': 'bcnThrBwMinBtoA',
        'bcnThresholdBandwRateAlertAtoB': 'bcnThrBwRtAlertAtoB',
        'bcnThresholdBandwRateAlertBtoA': 'bcnThrBwRtAlertBtoA',
        'bcnThresholdBandwRateWarningAtoB': 'bcnThrBwRtWarnAtoB',
        'bcnThresholdBandwRateWarningBtoA': 'bcnThrBwRtWarnBtoA',
        'bcnThresholdRRAlertAtoB': 'bcnThrRRAlertAtoB',
        'bcnThresholdRRAlertBtoA': 'bcnThrRRAlertBtoA',
        'bcnThresholdRRWarningAtoB': 'bcnThrRRWarnAtoB',
        'bcnThresholdRRWarningBtoA': 'bcnThrRRWarnBtoA',
        'bcnThresholdRTTAlertAtoB': 'bcnThrRTTAlertAtoB',
        'bcnThresholdRTTAlertBtoA': 'bcnThrRTTAlertBtoA',
        'bcnThresholdRTTWarningAtoB': 'bcnThrRTTWarnAtoB',
        'bcnThresholdRTTWarningBtoA': 'bcnThrRTTWarnBtoA',
        'bcnTrafficAtoB': 'bcnTrafficAtoB',
        'bcnTrafficBtoA': 'bcnTrafficBtoA',
    }
    perfdata_thresholds = {
        'bcnBandwidthAtoB': {
            'min': 'bcnThresholdBandwMinAtoB',
            'max': 'bcnThresholdBandwAvailableAtoB',
            'warn': 'bcnThresholdBandwRateWarningAtoB',
            'crit': 'bcnThresholdBandwRateAlertAtoB',
        },
        'bcnBandwidthBtoA': {
            'min': 'bcnThresholdBandwMinBtoA',
            'max': 'bcnThresholdBandwAvailableBtoA',
            'warn': 'bcnThresholdBandwRateWarningBtoA',
            'crit': 'bcnThresholdBandwRateAlertBtoA',
        },
        'bcnRTTAtoB': {
            'warn': 'bcnThresholdRTTWarningAtoB',
            'crit': 'bcnThresholdRTTAlertAtoB',
        },
        'bcnRTTBtoA': {
            'warn': 'bcnThresholdRTTWarningBtoA',
            'crit': 'bcnThresholdRTTAlertBtoA',
        },
        'bcnRRAtoB': {
            'warn': 'bcnThresholdRRWarningAtoB',
            'crit': 'bcnThresholdRRAlertAtoB',
        },
        'bcnRRBtoA': {
            'warn': 'bcnThresholdRRWarningBtoA',
            'crit': 'bcnThresholdRRAlertBtoA',
        },
    }
    perfdata_thresholds_percentages = ['bcnBandwidthAtoB', 'bcnBandwidthBtoA']

    @classmethod
    def get_label(cls, results):
        if options.bcnatob:
            way = '->'
        elif options.bcnbtoa:
            way = '<-'
        else:
            way = '<->'
        capture = results.get('bcnCaptureName', '')
        # result may be a literal 'None' from SNMP if no capture is set for the BCN
        if not capture or capture == 'None':
            capture = ''
        else:
            capture = ' (from %s)' % capture
        return "%s %s %s%s BCN %s" % (
            results['bcnZoneA'], way, results['bcnZoneB'], capture,
            cls.get_status(results))


if options.bca:
    mode = BCA
elif options.bcn:
    mode = BCN
else:
    assert False, "Invalid options"
# some nagios extensions may directly write the perfdata to rrd files
assert max(map(lambda x: len(x), mode.perfdata_aliases.values())) <= 19, 'broken RRD compatibility'


command, exit_code, stdout, stderr = snmp_walk([mode.table_oid])

if exit_code != 0:
    die("""snmpwalk failed.
To diagnose the problem, you may want to run %r yourself.""" % command)

# fetch the values corresponding to the query
id_ = None
results = {}
output_regex = re.compile(r'(?P<oid>\w+)\.(?P<id>\d+)\s+(?P<value>.*)')
for line in stdout.splitlines():
    match = output_regex.match(line)
    if match is None:
        continue
    match = match.groupdict()
    if mode.name_oid == match['oid'] and mode.query == match['value']:
        id_ = match['id']
    if id_ is not None and id_ == match['id']:
        results[match['oid']] = match['value']

# no results, either a typo or an invalid configuration
if not results:
    die("""No results found matching %r = %r.
Ensure the query is correct by running %r yourself.""" % (mode.name_oid, mode.query, command))


# display the status
sys.stdout.write(mode.get_label(results) + ' | ')
# followed by the performance data
for key in results:
    if key not in mode.perfdata_aliases:
        # display only selected perfdata
        continue

    # the perfdata value
    sys.stdout.write('%s=%s' % (mode.perfdata_aliases[key], results[key]))

    # optional thresholds for the perfdata
    thresholds = mode.get_thresholds(key, results)
    if thresholds:
        for threshold in ['warn', 'crit', 'min', 'max']: # order is important
            sys.stdout.write(';%s' % thresholds.get(threshold, ''))
    sys.stdout.write(' ')
sys.stdout.write('\n')

sys.exit(mode.get_exit_code(results))
